# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from typing import List, Optional

from contrast.agent.exclusions import TsExclusion, ExclusionType
from contrast.reporting.teamserver_responses.protect_rule import ProtectRule
from contrast.utils.decorators import cached_property


class ApplicationSettings:
    """
    This class is responsible for safely parsing V1 TeamServer Application Settings from a response to a usable format.
    The format can be found here: https://github.com/Contrast-Security-Inc/contrast-agent-api-spec. At the time of the
    creation of this class, the specific schema is ApplicationSettings in agent-endpoints.yml.
    """

    def __init__(self, application_settings_json: Optional[dict] = None):
        self.application_settings_json = application_settings_json or {}

    @cached_property
    def disabled_assess_rules(self) -> List[str]:
        return [
            rule_name
            for rule_name, rule_details in self.application_settings_json.get(
                "assess", {}
            ).items()
            if not rule_details.get("enable", False)
        ]

    @cached_property
    def exclusions(self) -> List[TsExclusion]:
        """
        Build and return a plain list of TsExclusions based on raw application settings. TsExclusions represent the
        exclusions more or less exactly as they were given to us by teamserver.

        The list returned by this property will be converted into our own Exclusions class later, which is more useful
        when processing exclusion rules.
        """
        if not (
            ts_exclusions := self.application_settings_json.get("exclusions", None)
        ):
            return []

        exclusions = []
        for exclusion in ts_exclusions.get("url", []):
            exclusions.append(
                TsExclusion(
                    ExclusionType.URL_EXCLUSION_TYPE,
                    exclusion.get("name", ""),
                    exclusion.get("modes", []),
                    exclusion.get("match_strategy", ""),
                    exclusion.get("protect_rules", []),
                    exclusion.get("assess_rules", []),
                    exclusion.get("urls", []),
                )
            )

        for exclusion in ts_exclusions.get("input", []):
            exclusions.append(
                TsExclusion(
                    ExclusionType.INPUT_EXCLUSION_TYPE,
                    # there's a hiccup in the API where "name" was repurposed for input exclusions
                    None,
                    exclusion.get("modes", []),
                    exclusion.get("match_strategy", ""),
                    exclusion.get("protect_rules", []),
                    exclusion.get("assess_rules", []),
                    exclusion.get("urls", []),
                    exclusion.get("type", ""),
                    # there's a hiccup in the API where "name" was repurposed for input exclusions
                    exclusion.get("name", ""),
                )
            )

        return exclusions

    @cached_property
    def protect_rules(self) -> List[ProtectRule]:
        return [
            ProtectRule({"id": name, **value})
            for name, value in self.application_settings_json.get("protect", {})
            .get("rules", {})
            .items()
            if value
        ]

    @cached_property
    def sensitive_data_masking_policy(self) -> dict:
        return self.application_settings_json.get("sensitive_data_masking_policy", {})

    @cached_property
    def session_id(self) -> Optional[str]:
        return self.application_settings_json.get("session_id", None)



# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from typing import Optional

from contrast.utils.decorators import cached_property


class Sampling:
    def __init__(self, sampling_json: dict, ng: bool):
        self.sampling_json = sampling_json or {}
        self.ng = ng

    @cached_property
    def enable(self) -> bool:
        if self.ng:
            return self.sampling_json.get("enabled", False)
        return self.sampling_json.get("enable", False)

    @cached_property
    def baseline(self) -> Optional[int]:
        return self.sampling_json.get("baseline", None)

    @cached_property
    def request_frequency(self) -> Optional[int]:
        if self.ng:
            return self.sampling_json.get("frequency", None)
        return self.sampling_json.get("request_frequency", None)

    @cached_property
    def window_ms(self) -> Optional[int]:
        if self.ng:
            if window := self.sampling_json.get("window", None):
                return window * 1000
            return None
        if window := self.sampling_json.get("window_ms", None):
            return window * 1000
        return None



# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import copy
from dataclasses import dataclass, field
from itertools import chain
from pathlib import PurePath
import re
import reprlib
import threading
import time
from typing import Any, Callable, Dict, List, Optional, Sequence, Type
from contrast.agent.assess.tag import Tag

from contrast.agent.policy.constants import (
    OBJECT,
    RETURN,
    TRIGGER_TYPE,
    CREATION_TYPE,
)

from contrast.agent.assess.truncate import truncate_tainted_string
from contrast.agent.assess.utils import get_properties

from contrast.agent.policy.policy_node import PolicyNode
from contrast.agent.settings import Settings
from contrast.api.trace_event import (
    TraceEvent,
    TraceEventObject,
    ParentObjectId,
    TraceEventSource,
)
from contrast.utils.assess.duck_utils import len_or_zero
from contrast.utils.base64_utils import B64_NONE_STRING, base64_encode
from contrast.utils.stack_trace_utils import (
    acceptable_frame,
    build_stack,
    clean_stack,
)
from contrast_vendor import structlog as logging
from contrast_vendor.webob.util import html_escape

logger = logging.getLogger("contrast")

INIT = "__init__"
INITIALIZERS = (INIT, "__new__")
NONE_STRING = str(None)


class Repr(reprlib.Repr):
    def repr_str(self, x, level):
        return x

    def repr_bytes(self, x, level):
        return x.decode(errors="replace")

    def repr_bytearray(self, x, level):
        return self.repr_bytes(x, level)

    def repr_instance(self, x, level):
        return f"{x.__class__.__module__}.{x.__class__.__qualname__}@{id(x):#x}"

    def repr_bool(self, x, level):
        return str(x)

    def repr_float(self, x, level):
        return str(x)

    def repr_NoneType(self, x, level):
        return "None"

    def repr_Pattern(self, x, level):
        return (
            self.repr1(x.pattern, level)
            if isinstance(x, re.Pattern)
            else self.repr_instance(x, level)
        )

    def repr_PurePath(self, x, level):
        return str(x) if isinstance(x, PurePath) else self.repr_instance(x, level)

    repr_PosixPath = repr_WindowsPath = repr_Path = repr_PurePosixPath = (
        repr_PureWindowsPath
    ) = repr_PurePath


class FullRepr(Repr):
    def repr_StringIO(self, x, level):
        return x.getvalue()

    def repr_BytesIO(self, x, level):
        return x.getvalue().decode(errors="replace")


class TruncatedRepr:
    def __init__(self, _repr: Repr):
        _repr = copy.copy(_repr)
        _repr.maxlevel = 3
        self._repr = _repr.repr

    def repr(self, obj: object):
        s = self._repr(obj)
        if len(s) > 103:
            s = s[:50] + "..." + s[-50:]
        return s


base_repr = Repr()
obj_repr = TruncatedRepr(base_repr)


def set_event_detail(detail: str):
    global obj_repr
    obj_repr = TruncatedRepr(base_repr if detail == "minimal" else FullRepr())


@dataclass
class Field:
    type: Type
    value: str

    def __init__(
        self, obj, typ=None, repr_func: Callable[[object], str] = obj_repr.repr
    ):
        self.type = typ or type(obj)
        self.value = repr_func(obj)


class ContrastEvent:
    """
    This class holds the data about an event in the application
    We'll use it to build an event that TeamServer can consume if
    the object to which this event belongs ends in a trigger.
    """

    ATOMIC_ID = 0

    def __init__(
        self,
        node: PolicyNode,
        tagged: Any,
        self_obj: Optional[Any],
        ret: Optional[Any],
        args: Sequence[Any],
        kwargs: Dict[str, Any],
        parents: List["ContrastEvent"],
        possible_key=None,
        source_type=None,
        source_name=None,
    ):
        self.node = node
        self._init_tagged(tagged, possible_key, args, kwargs)
        self.source_type = source_type
        self.source_name = source_name
        self.parents = parents or []
        ret = self._update_init_return(node, self_obj, ret)
        self.obj = Field(self_obj) if self_obj is not None else None
        self.ret = Field(ret) if ret is not None else None
        self.args = [Field(arg) for arg in args] if args is not None else None
        self.kwargs = (
            [Field(f"{k}={obj_repr.repr(v)}", typ=type(v)) for k, v in kwargs.items()]
            if kwargs is not None
            else None
        )

        # These are needed only at trigger-time but values must be set at init.
        self.time_ns = time.time_ns()
        self.thread = threading.current_thread().ident
        self.event_id = ContrastEvent._atomic_id()

        self.event_action = self.node.build_action()

        self._raw_stack = build_stack() if self._should_get_stack else []

        # This must happen at init for stream events to work.
        self._update_method_information()

    def _init_tagged(self, tagged: Any, possible_key=None, args=None, kwargs=None):
        """
        Initialize properties related to tagging.
        - self.tagged
        - self.taint_location
        - self.span_override
        """
        self.taint_location = possible_key or self._find_taint_location(args, kwargs)
        if self.taint_location and isinstance(tagged, dict):
            tagged = tagged.get(self.taint_location, None)
        tags = None
        self.tagged_props = get_properties(tagged)
        if self.tagged_props and isinstance(tagged, str):
            self.tagged_props.cleanup_tags()
            tags = [tag for tag in chain.from_iterable(self.tagged_props.tags.values())]
            tagged_repr, truncate_tags = truncate_tainted_string(tagged, tags)
            self.tagged = Field(tagged_repr, repr_func=base_repr.repr)
            self.tags_override = truncate_tags
        else:
            self.tagged = Field(tagged, repr_func=base_repr.repr)
            self.tags_override = (
                None
                if len(self.tagged.value) == len_or_zero(tagged)
                else (Tag(len(self.tagged.value), 0),)
            )

    @property
    def parent_ids(self):
        return [parent.event_id for parent in self.parents]

    @property
    def _should_get_stack(self):
        """
        Determine if event.stack should be populated or not.

        Get stacktrace for the event EXCEPT if
          1. the node explicitly indicates not to collect a stacktrace, which is the
                case for WSGI environ sources
          2. assess.stacktraces is configured to NONE
          3. assess.stacktraces is configured to SOME and event action is not
                creation or trigger


        :return: bool if to get stacktrace
        """
        settings = Settings()

        reportable_actions = (
            TraceEvent.Action[CREATION_TYPE],
            TraceEvent.Action[TRIGGER_TYPE],
        )

        if (
            self.node.skip_stacktrace
            or settings.is_collect_stacktraces_none()
            or (
                settings.is_collect_stacktraces_some()
                and self.event_action not in reportable_actions
            )
        ):
            return False

        return True

    @classmethod
    def _atomic_id(cls):
        ret = cls.ATOMIC_ID
        cls.ATOMIC_ID += 1
        return ret

    def _find_taint_location(self, args=None, kwargs=None):
        """
        Find the location of the tagged value in the event. This is used to determine
        where the tagged value is in the event, so that we can mark it up in TeamServer.
        """
        if len(self.node.targets) == 1:
            return self.node.targets[0]
        for loc in chain(self.node.targets, self.node.sources):
            if loc in (OBJECT, RETURN):
                return loc
            if isinstance(loc, int) and args and loc < len(args):
                return loc
            if isinstance(loc, str) and kwargs and loc in kwargs:
                return loc

        return None

    def _update_method_information(self):
        """
        For nicer reporting, we lie about the tagged value. For example, a call to
        split() returns a list of strings: ["foo", "bar"]. In the properties for "foo",
        the split event shows a return value of only "foo" instead of the whole list.
        """
        if self.taint_location is None:
            # This would be for trigger nodes without source or target. Trigger rule was
            # violated simply by a method being called. We'll save all the information,
            # but nothing will be marked up, as nothing need be tracked.
            return

        if self.taint_location == OBJECT:
            self.obj = self.tagged
            return

        if self.taint_location == RETURN:
            self.ret = self.tagged

    def to_reportable_event(self):
        """
        Convert a ContrastEvent to a TraceEvent.
        """
        event = TraceEvent()

        event.type = self.node.node_type
        event.action = self.event_action
        event.timestamp_ms = self.time_ns // 1_000_000
        event.thread = str(self.thread)
        event.tags = self.node.tags

        self._build_all_event_objects(event)

        event.stack = clean_stack(self._raw_stack, depth=20)

        safe_source_name = self.source_name if self.source_name is not None else ""
        event.field_name = safe_source_name

        if self.source_type:
            event.event_sources.append(self.build_source_event(safe_source_name))

        event.object_id = int(self.event_id)

        for parent_id in self.parent_ids:
            parent = ParentObjectId()
            parent.id = parent_id
            event.parent_object_ids.append(parent)

        self._build_complete_signature(event)
        self._validate_event(event)

        return event

    def _update_init_return(self, node, obj, ret):
        """
        For purposes of pretty reporting in Teamserver, we will say the
        `__init__` instance method return the `self` object (the object getting
        instantiated) instead of `None`, even though `None` is the return value of
        `__init__` methods.

        This will not apply if the node is not a class (it's possible someone
        creates a module level function called `__init__` or if the return
        value is already populated (for safety).
        """
        if node.method_name == INIT and node.class_name is not None:
            ret = ret or obj
        return ret

    def build_source_event(self, safe_source_name: str) -> TraceEventSource:
        """
        Create a new TraceEventSource
        """
        trace_event_source = TraceEventSource()

        trace_event_source.type = self.source_type
        trace_event_source.name = safe_source_name

        return trace_event_source

    def _set_event_source_and_target(self, event: TraceEvent):
        """
        We have to do a little work to figure out what our TS appropriate
        target is. To break this down, the logic is as follows:
        1) If my node has a target, work on targets. Else, work on sources.
           Per TS law, each node must have at least a source or a target.
           The only type of node w/o targets is a Trigger, but that may
           change.
        2) I'll set the event's source and target to TS values.
        """
        if self.node.targets:
            event.source = self.node.ts_valid_source
            event.target = self.node.ts_valid_target
        elif self.node.sources:
            event.source = self.node.ts_valid_target or self.node.ts_valid_source

    def _build_all_event_objects(self, event: TraceEvent):
        """
        First populate event.source and event.target
        Then populate fields of event.object, event.ret, and event.args (contains both
        args and kwargs for python) which are each `TraceEventObject`s
        """
        self._set_event_source_and_target(event)

        event.object.value = (
            base64_encode(self.obj.value) if self.obj else B64_NONE_STRING
        )
        event.ret.value = base64_encode(self.ret.value) if self.ret else B64_NONE_STRING
        if self.args:
            event.args = [
                TraceEventObject(arg.value if arg else B64_NONE_STRING)
                for arg in self.args
            ]
        if self.kwargs:
            event_kwargs = [
                TraceEventObject(kwarg.value if kwarg else B64_NONE_STRING)
                for kwarg in self.kwargs
            ]
            event.args.extend(event_kwargs)

        self._add_taint_ranges(event, self.taint_location)

    def _add_taint_ranges(self, event: TraceEvent, taint_location: str):
        """
        Populate event.taint_ranges
        """
        if taint_location is None or not self.tagged_props:
            return

        event.taint_ranges = self.tagged_props.tags_to_ts_obj(self.tags_override)

        # For now, only the taint_target needs to be officially marked as tracked.
        # This means that agent-tagged strings may not be marked as "tracked" for TS.
        # This may change in the future if we change the corresponding TS endpoint; in
        # that case, use recursive_is_tracked for each TraceEventObject.
        if taint_location == OBJECT:
            event_obj = event.object
        elif taint_location == RETURN:
            event_obj = event.ret
        elif isinstance(taint_location, int):
            event_obj = event.args[taint_location]
        elif isinstance(taint_location, str):
            event_obj = event.args[-1]

        event_obj.tracked = True

    def _build_complete_signature(self, event: TraceEvent):
        return_type = self.ret.type.__name__ if self.ret is not None else NONE_STRING

        event.signature.return_type = return_type
        # We don't want to report "BUILTIN" as a module name in Team Server
        event.signature.class_name = self.node.location.replace("BUILTIN.", "")
        event.signature.method_name = self.node.method_name

        if self.args:
            event.signature.arg_types.extend(
                [
                    arg.type.__name__ if arg is not None else NONE_STRING
                    for arg in self.args
                ]
            )

        if self.kwargs:
            event.signature.arg_types.extend(
                [
                    kwarg.type.__name__ if kwarg is not None else NONE_STRING
                    for kwarg in self.kwargs
                ]
            )

        event.signature.constructor = self.node.method_name in INITIALIZERS

        # python always returns None if not returned
        event.signature.void_method = False

        if not self.node.instance_method:
            event.signature.flags = 8

    def _validate_event(self, event: TraceEvent):
        """
        TS is not able to render a vulnerability correctly if the source string index 0
        of the trigger event, ie event.source, is not a known one.

        See TS repo DataFlowSnippetBuilderVersion1.java:buildMarkup

        :param event: TraceEvent
        :return: None
        """
        allowed_trigger_sources = ["O", "P", "R"]
        if (
            event.action == TraceEvent.Action[TRIGGER_TYPE]
            and event.source[0] not in allowed_trigger_sources
        ):
            # If this is logged, check the node in policy.json corresponding to
            # this event and how the agent has transformed the source string
            logger.debug("WARNING: trigger event TS-invalid source %s", event.source)

    def history(self: "ContrastEvent"):
        """
        Return a generator that yields all the events in the history of this event,
        starting with the event itself and then its parents, and so on.

        Events are yielded in depth-first order, so the event itself is yielded first,
        then its last parent, then the last parent of that parent, and so on.

        Events are deduplicated, so if an event is seen more than once, it will not be
        yielded again.
        """
        seen = set()
        queue = [self]
        while queue:
            event = queue.pop()
            seen.add(event)
            yield event
            queue.extend(event for event in event.parents if event not in seen)

    def dot_repr(self: "ContrastEvent") -> str:
        """
        Returns a DOT graph representation of self's history.
        """
        return str(DotGraph(self))


@dataclass
class DotGraph:
    event: ContrastEvent
    normalize: bool = False
    _seen_events: Dict[ContrastEvent, int] = field(default_factory=dict, init=False)

    def __str__(self):
        dot_lines = [
            "digraph {",
            "   node [shape=plain];",
        ]
        for event in self.event.history():
            dot_lines.append(self._node(event))
            dot_lines.extend(self._edge(parent, event) for parent in event.parents)
        dot_lines.append("}")
        return "\n".join(dot_lines)

    def _node(self, event: ContrastEvent) -> str:
        return f"""{self._event_id(event)} [label=<
        <table cellborder="0" cellspacing="10" style="rounded" {self._tooltip(event)}>
        <tr><td align="text" cellpadding="1"><b>{event.node}</b></td></tr>
        <tr><td><table cellborder="1" cellspacing="0" cellpadding="15">
            <tr><td align="text">data</td><td align="text"><font face="Monospace">{html_escape(event.tagged.value)}</font></td></tr>
            {"".join(f"<tr><td align='text'>{tag}</td><td align='text'><font face='Monospace'>{self._tagrng_markup(rngs, len(event.tagged.value), untagged_marker='&nbsp;')}</font></td></tr>" for tag, rngs in (event.tagged_props.tags.items() if event.tagged_props else []) )}
        </table></td></tr>
        </table>>];"""

    def _edge(self, parent: ContrastEvent, child: ContrastEvent) -> str:
        return f"{self._event_id(parent)} -> {self._event_id(child)};"

    def _event_id(self, event: ContrastEvent) -> int:
        if self.normalize:
            if event not in self._seen_events:
                self._seen_events[event] = len(self._seen_events) + 1
            return self._seen_events[event]

        return event.event_id

    def _tooltip(self, event: ContrastEvent) -> str:
        if self.normalize:
            return ""

        # intentionally avoid using clean_stack because it formats filenames
        # replacing / with ., which makes them unusable as links.
        frame = next(
            (frame for frame in reversed(event._raw_stack) if acceptable_frame(frame)),
            None,
        )

        return (
            f'tooltip="{html_escape(frame.line)}" href="file://{frame.filename}"'
            if frame
            else ""
        )

    def _tagrng_markup(self, rngs, length, tag_marker="*", untagged_marker=" ") -> str:
        return "".join(
            tag_marker if any(rng.covers(i) for rng in rngs) else untagged_marker
            for i in range(length)
        )



# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Using logger.info/debug/someOtherLevel() is not supported in this module. In order to get the correct
frame info, we must skip over functions called in this module and in vendored structlog. If logging is attempted,
incorrect frame info will be displayed on the log message if used in this file.

Use print(...) instead
"""
import asyncio

from contrast.agent import request_state
from contrast_vendor import structlog
from contrast.utils.configuration_utils import get_hostname

LOGGING_TO_BUNYAN_LOG_LEVEL_CONVERSION = {
    "critical": 60,
    "error": 50,
    "warning": 40,
    "info": 30,
    "debug": 20,
}


# TODO: PYT-3186 there should be no references to structlog.stdlib after this ticket is
# completed


def add_hostname_for_bunyan(logger, method_name, event_dict):
    event_dict["hostname"] = get_hostname()
    return event_dict


def add_request_id(logger, method_name, event_dict):
    event_dict["request_id"] = request_state.get_request_id()
    return event_dict


def rename_key_for_bunyan(old_name, new_name):
    def key_renamer(logger, method_name, event_dict):
        value = event_dict.get(old_name)
        if value and not event_dict.get(new_name):
            event_dict[new_name] = value
            del event_dict[old_name]

        return event_dict

    return key_renamer


def add_log_level_for_bunyan(logger, log_level, event_dict):
    """
    NOTE: Bunyan uses different log level integers than logging/structlog.
    """
    if log_level == "warn":
        # The stdlib has an alias
        log_level = "warning"

    new_value = LOGGING_TO_BUNYAN_LOG_LEVEL_CONVERSION.get(log_level, None)

    if new_value:
        event_dict["level"] = new_value

    return event_dict


def add_v_for_bunyan(logger, method_name, event_dict):
    event_dict["v"] = 0
    return event_dict


def add_progname_for_bunyan(logger, method_name, event_dict):
    """
    progname is the name of the process the agents uses in logs.
    The default value is Contrast Agent. progname will be used
    as the name of the logger as seen in the logs.
    """
    field = "name"
    current_handler = logger.handlers[0]

    if hasattr(current_handler.filters[0], field):
        progname = current_handler.filters[0].progname

        if progname:
            event_dict[field] = progname

    return event_dict


def add_asyncio_info(logger, method_name, event_dict):
    try:
        current_task = asyncio.current_task()

        # If no name has been explicitly assigned to the Task, the default asyncio Task implementation
        # generates a default name during instantiation.
        event_dict["asyncio_task_name"] = current_task.get_name()

        current_coro = current_task.get_coro()
        if hasattr(current_coro, "__name__"):
            event_dict["asyncio_coro_name"] = current_coro.__name__

        event_dict["asyncio_task_id"] = id(current_task)
    except Exception:
        # This can happen when there is no running event loop
        pass

    return event_dict


def init_structlog():
    """
    Initial configuration for structlog. This can still be modified by subsequent calls
    to structlog.configure.
    """
    structlog.configure(
        processors=[
            # TODO: PYT-3186 stop filtering here once we're using a filtering bound logger
            structlog.stdlib.filter_by_level,
            # exact key "time" required for bunyan parsing
            structlog.processors.TimeStamper(fmt="iso", key="time"),
            structlog.stdlib.PositionalArgumentsFormatter(),
            add_log_level_for_bunyan,
            rename_key_for_bunyan("event", "msg"),
            structlog.processors.CallsiteParameterAdder(
                [
                    structlog.processors.CallsiteParameter.PROCESS,
                    structlog.processors.CallsiteParameter.FILENAME,
                    structlog.processors.CallsiteParameter.FUNC_NAME,
                    structlog.processors.CallsiteParameter.LINENO,
                    structlog.processors.CallsiteParameter.THREAD,
                ],
                additional_ignores=[
                    "contrast_vendor.structlog",
                    "contrast.utils.decorators",
                ],
            ),
            rename_key_for_bunyan("process", "pid"),
            add_request_id,
            add_asyncio_info,
            add_hostname_for_bunyan,
            add_progname_for_bunyan,
            add_v_for_bunyan,
            structlog.processors.format_exc_info,
            structlog.processors.StackInfoRenderer(
                additional_ignores=[
                    "contrast_vendor.structlog",
                    "contrast.utils.decorators",
                ]
            ),
            structlog.processors.UnicodeDecoder(),
            structlog.processors.JSONRenderer(),
        ],
        # TODO: PYT-3186 this will need to be something like PrintLoggerFactory()
        # or WriteLoggerFactory(). It also may need to change later.
        logger_factory=structlog.stdlib.LoggerFactory(),
        # TODO: PYT-3186 use this wrapper class instead of one from structlib.stdlib
        # wrapper_class=structlog.make_filtering_bound_logger(log_level),
        wrapper_class=structlog.stdlib.BoundLogger,
        # TODO: PYT-3168 it's possible that using this option will prevent us from
        # being able to modify the logger on subsequent calls to structlog.configure.
        # If that happens, we need to make sure that not every logging call recreates
        # and configures a new logger instance.
        cache_logger_on_first_use=True,
    )



# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.


CONTRAST_SCOPE: int
PROPAGATION_SCOPE: int
TRIGGER_SCOPE: int


def in_contrast_or_propagation_scope() -> bool:
    ...


def has_funchook() -> bool:
    ...


def get_current_scope() -> tuple:
    ...


def enter_scope(scope: int, /) -> None:
    ...


def exit_scope(scope: int, /) -> None:
    ...


def in_scope(scope: int, /) -> bool:
    ...


def set_exact_scope(scope: tuple[int, int, int], /) -> None:
    ...


def get_tp_version_tag(type: type, /) -> int:
    ...


def set_attr_on_type(owner: type, name: str, value, /) -> None:
    ...



# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Controller for global scope state

Basically we use scoping to prevent us from assessing our own code. Scope
improves performance but it also prevents us from accidentally recursing
inside our analysis code. For example, we don't want to inadvertently cause
string propagation events while we're doing string building for reporting
purposes.
"""
from typing import Callable, Generator

import functools
import contextlib

from contrast.assess_extensions import cs_str


def enter_contrast_scope() -> None:
    """
    Enter contrast scope

    Contrast scope is global. It should prevent us from taking *any*
    further analysis action, whether it be propagation or evaluating
    triggers.
    """
    cs_str.enter_scope(cs_str.CONTRAST_SCOPE)


def enter_propagation_scope() -> None:
    """
    Enter propagation scope

    While in propagation scope, prevent any further propagation actions.
    Basically this means that no string propagation should occur while in
    propagation scope.
    """
    cs_str.enter_scope(cs_str.PROPAGATION_SCOPE)


def enter_trigger_scope() -> None:
    """
    Enter trigger scope

    While in trigger scope, prevent analysis inside of any other trigger
    methods that get called.
    """
    cs_str.enter_scope(cs_str.TRIGGER_SCOPE)


def exit_contrast_scope() -> None:
    cs_str.exit_scope(cs_str.CONTRAST_SCOPE)


def exit_propagation_scope() -> None:
    cs_str.exit_scope(cs_str.PROPAGATION_SCOPE)


def exit_trigger_scope() -> None:
    cs_str.exit_scope(cs_str.TRIGGER_SCOPE)


def in_contrast_scope() -> bool:
    return cs_str.in_scope(cs_str.CONTRAST_SCOPE)


def in_propagation_scope() -> bool:
    return cs_str.in_scope(cs_str.PROPAGATION_SCOPE)


def in_trigger_scope() -> bool:
    return cs_str.in_scope(cs_str.TRIGGER_SCOPE)


@contextlib.contextmanager
def contrast_scope() -> Generator[None, None, None]:
    cs_str.enter_scope(cs_str.CONTRAST_SCOPE)
    try:
        yield
    finally:
        cs_str.exit_scope(cs_str.CONTRAST_SCOPE)


@contextlib.contextmanager
def propagation_scope() -> Generator[None, None, None]:
    cs_str.enter_scope(cs_str.PROPAGATION_SCOPE)
    try:
        yield
    finally:
        cs_str.exit_scope(cs_str.PROPAGATION_SCOPE)


@contextlib.contextmanager
def trigger_scope() -> Generator[None, None, None]:
    cs_str.enter_scope(cs_str.TRIGGER_SCOPE)
    try:
        yield
    finally:
        cs_str.exit_scope(cs_str.TRIGGER_SCOPE)


def with_contrast_scope(orig_func: Callable) -> Callable:
    @functools.wraps(orig_func)
    def contrast_scope_wrapper(*args, **kwargs):
        cs_str.enter_scope(cs_str.CONTRAST_SCOPE)
        try:
            return orig_func(*args, **kwargs)
        finally:
            cs_str.exit_scope(cs_str.CONTRAST_SCOPE)

    return contrast_scope_wrapper


def with_propagation_scope(orig_func: Callable) -> Callable:
    @functools.wraps(orig_func)
    def propagation_scope_wrapper(*args, **kwargs):
        cs_str.enter_scope(cs_str.PROPAGATION_SCOPE)
        try:
            return orig_func(*args, **kwargs)
        finally:
            cs_str.exit_scope(cs_str.PROPAGATION_SCOPE)

    return propagation_scope_wrapper


def with_trigger_scope(orig_func: Callable) -> Callable:
    @functools.wraps(orig_func)
    def trigger_scope_wrapper(*args, **kwargs):
        cs_str.enter_scope(cs_str.TRIGGER_SCOPE)
        try:
            return orig_func(*args, **kwargs)
        finally:
            cs_str.exit_scope(cs_str.TRIGGER_SCOPE)

    return trigger_scope_wrapper


###################################################
# Convenience functions not needed for all scopes #
###################################################


def in_contrast_or_propagation_scope():
    """Indicates we are in either contrast scope or propagation scope"""
    return cs_str.in_contrast_or_propagation_scope()


@contextlib.contextmanager
def pop_contrast_scope():
    """
    Context manager that pops contrast scope level and restores it when it exits

    Scope is implemented as a stack. If the thread is in contrast scope at the time
    this is called, the scope level will be reduced by one for the lifetime of the
    context manager. If the prior scope level was 1, this has the effect of temporarily
    disabling contrast scope. The original scope level will be restored when the
    context manager exits. If the thread is **not** already in contrast scope when this
    is called, it has no effect.
    """
    was_in_scope = cs_str.in_scope(cs_str.CONTRAST_SCOPE)
    # This has no effect if we're not already in scope
    cs_str.exit_scope(cs_str.CONTRAST_SCOPE)
    try:
        yield
    finally:
        # For safety, only restore scope if we were in it to begin with
        if was_in_scope:
            cs_str.enter_scope(cs_str.CONTRAST_SCOPE)



# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys
import traceback
from functools import lru_cache

from contrast import AGENT_CURR_WORKING_DIR
from contrast.api.trace_event import TraceStack
from contrast.utils.decorators import fail_quietly
from contrast.utils.string_utils import removeprefix
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


APPLIES_MARKER = "cs__"
PATCH_MARKER = "__cs"
PY_FILE_EXTENSION = ".py"
UTILS_MODULES = "contrast/utils"
NONETYPE = "NoneType"

CONTRAST_EXTENSIONS = (
    "contrast_vendor",
    "contrast_rewriter",
    "contrast/assess_extensions",
    "contrast/patches",
)
DJANGO_EXCEPTION_PATH = "core/handlers/exception.py"
DJANGO_DEPRECATION_PATH = "utils/deprecation.py"
SITE_PACKAGES = "site-packages"


@fail_quietly("Failed to build stacktrace for event", return_value=[])
def build_stack():
    # PERF: only grab 30 elements to not use as much memory.
    # But the clean step will likely remove more elements.
    return traceback.extract_stack(limit=30)


@fail_quietly("Failed to clean stacktrace for event", return_value=[])
def clean_stack(frames, depth=20):
    frames.reverse()  # moves most recent to the front

    frames = [frame for frame in frames if acceptable_frame(frame)]

    return to_element_list(frames[:depth])


def acceptable_frame(frame):
    """
    Return true if frame does NOT contain contrast or __cs
    """
    ignore_string = "/contrast/"

    if isinstance(frame, tuple):
        return (
            ignore_string not in frame[0]
            and UTILS_MODULES not in frame[0]
            and not any(extension in frame[0] for extension in CONTRAST_EXTENSIONS)
            and not frame[2].startswith(APPLIES_MARKER)
            and not frame[2].startswith(PATCH_MARKER)
            and not frame[0].endswith(DJANGO_EXCEPTION_PATH)
            and not frame[0].endswith(DJANGO_DEPRECATION_PATH)
        )

    return (
        ignore_string not in frame.filename
        and UTILS_MODULES not in frame.filename
        and not any(extension in frame.filename for extension in CONTRAST_EXTENSIONS)
        and not frame.name.startswith(APPLIES_MARKER)
        and not frame.name.startswith(PATCH_MARKER)
        and not frame.filename.endswith(DJANGO_EXCEPTION_PATH)
        and not frame.filename.endswith(DJANGO_DEPRECATION_PATH)
    )


def build_and_clean_stack(depth=10):
    """
    Perform both build and clean steps.
    """
    frames = build_stack()
    return clean_stack(frames, depth=depth)


def to_element_list(frames):
    return [y for y in [to_element(x) for x in frames] if y is not None]


@fail_quietly("Failed to convert event")
def to_element(summary):
    if not summary:
        return None

    if isinstance(summary, tuple):
        # in python 2 traceback returns a tuple
        path = summary[0]
        method = summary[2]
        line_number = summary[1]
    else:
        # python 3 is a FrameSummary
        path = summary.filename
        method = summary.name
        line_number = summary.lineno

    element = TraceStack()
    element.line_number = line_number

    element.file_name = filename_formatter(path)
    element.declaring_class = element.file_name
    element.method_name = method

    return element


SORTED_FILENAME_SEARCH_PATH = sorted(
    set(sys.path) | {AGENT_CURR_WORKING_DIR}, key=len, reverse=True
)


@fail_quietly("Unable to create file_name")
@lru_cache(maxsize=512)
def filename_formatter(file_name: str):
    # PERF: This method is called hundreds of times, so be mindful
    # of what additional computations are added.

    if file_name.startswith("<frozen"):
        return file_name

    for sys_path in SORTED_FILENAME_SEARCH_PATH:
        if file_name.startswith(sys_path):
            file_name = removeprefix(file_name, sys_path)
            break

    return file_name.replace("/", ".").lstrip(".")



# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from typing import Optional

from contrast.agent import agent_state
from contrast.agent.middlewares.route_coverage import common

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class RoutesMixin:
    """
    Class for route coverage work.
    Route coverage is assess only.
    """

    def handle_routes(self, context, request) -> None:
        """
        Method that should run for all middlewares at the end of request processing.
        """
        if context and request:
            self.update_route_information(context, request)
            self.append_route_to_findings(context)

    def update_route_information(self, context, request) -> None:
        """
        Store the information for this request's route on RequestContext.
        """
        logger.debug("updating observed route information")

        signature = self._get_signature(context, request)

        context.observed_route.signature = signature
        context.observed_route.url = context.request.get_normalized_uri()
        context.observed_route.verb = request.method
        logger.debug(
            "observed route",
            signature=context.observed_route.signature,
            url=context.observed_route.url,
            verb=context.observed_route.verb,
        )

    def _get_signature(self, context, request) -> Optional[str]:
        """
        The logic here is somewhat convoluted and depends on state being set in a few
        other locations. Much of this will change as we move route discovery to a
        background thread. There are a few different strategies we might use to obtain a
        route signature. In order of preference:

        1.  Obtain a signature at some point during the request; this string is set on
            `context.view_func_str`. For supported frameworks, this is accomplished with
            framework-specific patches. One exception to this is aiohttp, where it
            happens directly in the middleware. It is essential that for a particular
            route this signature exactly matches the one found during route discovery.
            In the future, all frameworks will take this approach.

        2.  Obtain a reference to the view function at some point during the request.
            This reference is set on `context.view_func`.
            2a. Use this reference to construct a key to the discovered_routes dict.
                Look up the previously discovered route, and use its signature. An
                unfortunate consequence of this approach is that route discovery must be
                completed first, so we can't do this anymore if we want to move route
                discovery to a background thread.
            2b. If we don't find the corresponding discovered route, build a reasonable
                signature using the view function we have. This is a "dynamic route".

        3.  If neither context.view_func_str nor context.view_func are set, we didn't
            get any framework-specific information about the view function. In this case
            we use the request's normalized URI as the signature. We expect to hit this
            case in pure WSGI / ASGI.

        Framework-specific route observation code should set either view_func_str, which
        takes priority, or view_func. It doesn't make sense to set both, since only
        view_func_str would be used in that case.
        """
        if context.view_func_str:
            return context.view_func_str

        # TODO: PYT-3317 remove this entire block
        if context.view_func:
            discovered_routes = agent_state.get_discovered_routes()
            key = common.build_key(str(id(context.view_func)), request.method)
            if route := discovered_routes.get(key):
                return route.signature
            return common.build_route(context.view_func.__name__, context.view_func)

        logger.debug(
            "Did not find a view function signature for the current request. "
            "Falling back on normalized URI."
        )
        return context.request.get_normalized_uri()

    def append_route_to_findings(self, context):
        """
        Append the observed route to any existing findings. We can't necessarily do this
        at finding creation time, because we might not have route info yet.
        """
        if not context.observed_route:
            logger.debug("No observed route to append to findings")
            return

        if not context.findings:
            logger.debug("No findings to append route to")
            return

        for finding in context.findings:
            if not finding.routes:
                logger.debug(
                    "Appending route %s:%s to %s",
                    context.observed_route.verb,
                    context.observed_route.url,
                    finding.rule_id,
                )
                finding.routes.append(context.observed_route)

            finding.set_version()



# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from typing import Optional

from contrast.reporting.teamserver_responses.sampling import Sampling
from contrast.utils.decorators import cached_property


class NGServerSettings:
    def __init__(self, features: Optional[dict] = None):
        self.features = features or {}

    @cached_property
    def log_file(self) -> str:
        return self.features.get("features", {}).get("logFile", "")

    @cached_property
    def log_level(self) -> str:
        return self.features.get("features", {}).get("logLevel", "")

    @cached_property
    def assess_enabled(self) -> bool:
        return (
            self.features.get("features", {})
            .get("assessment", {})
            .get("enabled", False)
        )

    @cached_property
    def protect_enabled(self) -> bool:
        return self.features.get("features", {}).get("defend", {}).get("enabled", False)

    @cached_property
    def sampling(self) -> Sampling:
        return Sampling(
            self.features.get("features", {})
            .get("assessment", {})
            .get("sampling", None),
            True,
        )



# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from typing import List, Optional

from contrast.agent.exclusions import TsExclusion, ExclusionType
from contrast.reporting.teamserver_responses.protect_rule import ProtectRule
from contrast.utils.decorators import cached_property


class NGApplicationSettings:
    def __init__(self, settings_json=None):
        self._raw_settings = settings_json or {}

    @cached_property
    def disabled_assess_rules(self) -> List[str]:
        return (
            self._raw_settings.get("settings", {})
            .get("assessment", {})
            .get("disabledRules", [])
        )

    @cached_property
    def protection_rules(self) -> List[ProtectRule]:
        return [
            ProtectRule(r)
            for r in self._raw_settings.get("settings", {})
            .get("defend", {})
            .get("protectionRules", [])
        ]

    @cached_property
    def session_id(self) -> Optional[str]:
        return (
            self._raw_settings.get("settings", {})
            .get("assessment", {})
            .get("session_id", None)
        )

    @cached_property
    def sensitive_data_masking_policy(self) -> dict:
        return self._raw_settings.get("settings", {}).get(
            "sensitive_data_masking_policy"
        )

    @cached_property
    def exclusions(self) -> Optional[List[TsExclusion]]:
        """
        Build and return a plain list of TsExclusions based on raw application settings.
        TsExclusions represent the exclusions more or less exactly as they were given to
        us by teamserver.

        The list returned by this property will be converted into our own Exclusions
        class later, which is more useful when processing exclusion rules.
        """
        ts_exclusions = self._raw_settings.get("settings", {}).get("exceptions")
        if ts_exclusions is None:
            return None

        exclusions = []

        for exclusion in ts_exclusions.get("urlExceptions", []):
            exclusions.append(
                TsExclusion(
                    ExclusionType.URL_EXCLUSION_TYPE,
                    exclusion.get("name", ""),
                    exclusion.get("modes", []),
                    exclusion.get("matchStrategy", ""),
                    exclusion.get("protectionRules", []),
                    exclusion.get("assessmentRules", []),
                    exclusion.get("urls", []),
                )
            )

        for exclusion in ts_exclusions.get("inputExceptions", []):
            exclusions.append(
                TsExclusion(
                    ExclusionType.INPUT_EXCLUSION_TYPE,
                    exclusion.get("name", ""),
                    exclusion.get("modes", []),
                    exclusion.get("matchStrategy", ""),
                    exclusion.get("protectionRules", []),
                    exclusion.get("assessmentRules", []),
                    exclusion.get("urls", []),
                    exclusion.get("inputType", ""),
                    exclusion.get("inputName", ""),
                )
            )

        return exclusions

    # note: there are more fields on ApplicationSettings that we currently don't use



# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import requests

from .base_ts_message import BaseTsAppMessage
from .effective_config import EffectiveConfig
from contrast.reporting.activity_masker import ActivityMasker
from contrast.utils.decorators import fail_loudly
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class ApplicationActivity(BaseTsAppMessage):
    def __init__(self, context):
        super().__init__()

        self.body = {"lastUpdate": self.since_last_update}

        self.body["inventory"] = {
            # Used by TeamServer to aggregate counts across a given time period, for
            # Protect and attacker activity.
            "activityDuration": 0,
            "components": [],
        }
        if context.request and context.request.user_agent:
            self.body["inventory"]["browsers"] = [context.request.user_agent]
        if context.database_info:
            for db_info in context.database_info:
                self.body["inventory"]["components"].append(db_info)

        if context.attacks:
            # The only sensitive data in this message is the request field
            # of samples. Since Assess messages won't have samples, we
            # delay masking until this point.
            ActivityMasker().mask_sensitive_data(context.request, context.attacks)

            self.body["defend"] = {"attackers": []}

            for attack in context.attacks:
                self.body["defend"]["attackers"].append(
                    {
                        "protectionRules": {
                            attack.rule_id: attack.to_json(context.request)
                        },
                        "source": {
                            "ip": context.request.client_addr or "",
                            "xForwardedFor": context.request.headers.get(
                                "X-Forwarded-For"
                            )
                            or "",
                        },
                    }
                )

    @property
    def name(self):
        return "application-activity"

    @property
    def path(self):
        return "activity/application"

    @property
    def request_method(self):
        return requests.Session.put

    @property
    def expected_response_codes(self):
        return [200, 204]

    @fail_loudly("Failed to process Activity response")
    def process_response(self, response, reporting_client):
        if not self.process_response_code(response, reporting_client):
            return

        body = response.json()

        self.settings.apply_ng_app_settings(body)
        self.settings.process_ts_reactions(body)
        self.settings.log_effective_config()
        if reporting_client is not None and self.settings.is_agent_config_enabled():
            reporting_client.add_message(EffectiveConfig())



# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from typing import Optional

from contrast.reporting.teamserver_responses.sampling import Sampling
from contrast.utils.decorators import cached_property


class ServerSettings:
    """
    This class is responsible for safely parsing V1 TeamServer Server Settings from a response to a usable format.
    The format can be found here: https://github.com/Contrast-Security-Inc/contrast-agent-api-spec. At the time of the
    creation of this class, the specific schema is ServerSettings in agent-endpoints.yml.
    """

    def __init__(self, server_settings_json: Optional[dict] = None):
        self.server_settings_json = server_settings_json or {}

    @cached_property
    def log_file(self) -> Optional[str]:
        return self.server_settings_json.get("logger", {}).get("path", None)

    @cached_property
    def log_level(self) -> Optional[str]:
        return self.server_settings_json.get("logger", {}).get("level", None)

    @cached_property
    def assess_enabled(self) -> bool:
        return self.server_settings_json.get("assess", {}).get("enable", False)

    @cached_property
    def protect_enabled(self) -> bool:
        return self.server_settings_json.get("protect", {}).get("enable", False)

    @cached_property
    def sampling(self) -> Optional[Sampling]:
        return Sampling(
            self.server_settings_json.get("assess", {}).get("sampling", None), False
        )
